@@ -59,6 +59,204 @@ define([
     /** @var {Object} iconSystem - Object extending core/iconsystem */
     var iconSystem = {};
 
+    /** @var {Object[]} loadTemplateBuffer - List of templates to be loaded */
+    var loadTemplateBuffer = [];
+
+    /** @var {Bool} isLoadingTemplates - Whether templates are currently being loaded */
+    var isLoadingTemplates = false;
+
+    /** @var {Array} blacklistedNestedHelpers - List of helpers that can't be called within other helpers */
+    var blacklistedNestedHelpers = ['js'];
+
+    /**
+     * Search the various caches for a template promise for the given search key.
+     * The search key should be in the format <theme>/<component>/<template> e.g. boost/core/modal.
+     *
+     * If the template is found in any of the caches it will populate the other caches with
+     * the same data as well.
+     *
+     * @param {String} searchKey The template search key in the format <theme>/<component>/<template> e.g. boost/core/modal
+     * @return {Object} jQuery promise resolved with the template source
+     */
+    var getTemplatePromiseFromCache = function(searchKey) {
+        // First try the cache of promises.
+        if (searchKey in templatePromises) {
+            return templatePromises[searchKey];
+        }
+
+        // Check the module cache.
+        if (searchKey in templateCache) {
+            // Add this to the promises cache for future.
+            templatePromises[searchKey] = $.Deferred().resolve(templateCache[searchKey]).promise();
+            return templatePromises[searchKey];
+        }
+
+        if (M.cfg.templaterev <= 0) {
+            // Template caching is disabled. Do not store in persistent storage.
+            return null;
+        }
+
+        // Now try local storage.
+        var cached = storage.get('core_template/' + M.cfg.templaterev + ':' + searchKey);
+        if (cached) {
+            // Add this to the module cache for future.
+            templateCache[searchKey] = cached;
+            // Add this to the promises cache for future.
+            templatePromises[searchKey] = $.Deferred().resolve(cached).promise();
+            return templatePromises[searchKey];
+        }
+
+        return null;
+    };
+
+    /**
+     * Take all of the templates waiting in the buffer and load them from the server
+     * or from the cache.
+     *
+     * All of the templates that need to be loaded from the server will be batched up
+     * and sent in a single network request.
+     */
+    var processLoadTemplateBuffer = function() {
+        if (!loadTemplateBuffer.length) {
+            return;
+        }
+
+        if (isLoadingTemplates) {
+            return;
+        }
+
+        isLoadingTemplates = true;
+        // Grab any templates waiting in the buffer.
+        var templatesToLoad = loadTemplateBuffer.slice();
+        // This will be resolved with the list of promises for the server request.
+        var serverRequestsDeferred = $.Deferred();
+        var requests = [];
+        // Get a list of promises for each of the templates we need to load.
+        var templatePromises = templatesToLoad.map(function(templateData) {
+            var component = templateData.component;
+            var name = templateData.name;
+            var searchKey = templateData.searchKey;
+            var theme = templateData.theme;
+            var templateDeferred = templateData.deferred;
+            var promise = null;
+
+            // Double check to see if this template happened to have landed in the
+            // cache as a dependency of an earlier template.
+            var cachedPromise = getTemplatePromiseFromCache(searchKey);
+            if (cachedPromise) {
+                // We've seen this template so immediately resolve the existing promise.
+                promise = cachedPromise;
+            } else {
+                // We haven't seen this template yet so we need to request it from
+                // the server.
+                requests.push({
+                    methodname: 'core_output_load_template_with_dependencies',
+                    args: {
+                        component: component,
+                        template: name,
+                        themename: theme,
+                        lang: $('html').attr('lang').replace(/-/g, '_')
+                    }
+                });
+                // Remember the index in the requests list for this template so that
+                // we can get the appropriate promise back.
+                var index = requests.length - 1;
+
+                // The server deferred will be resolved with a list of all of the promises
+                // that were sent in the order that they were added to the requests array.
+                promise = serverRequestsDeferred.promise()
+                    .then(function(promises) {
+                        // The promise for this template will be the one that matches the index
+                        // for it's entry in the requests array.
+                        //
+                        // Make sure the promise is added to the promises cache for this template
+                        // search key so that we don't request it again.
+                        templatePromises[searchKey] = promises[index].then(function(response) {
+                            var templateSource = null;
+
+                            // Process all of the template dependencies for this template and add
+                            // them to the caches so that we don't request them again later.
+                            response.templates.forEach(function(data) {
+                                // Generate the search key for this template in the response so that we
+                                // can add it to the caches.
+                                var tempSearchKey = [theme, data.component, data.name].join('/');
+                                // Cache all of the dependent templates because we'll need them to render
+                                // the requested template.
+                                templateCache[tempSearchKey] = data.value;
+
+                                if (M.cfg.templaterev > 0) {
+                                    // The template cache is enabled - set the value there.
+                                    storage.set('core_template/' + M.cfg.templaterev + ':' + tempSearchKey, data.value);
+                                }
+
+                                if (data.component == component && data.name == name) {
+                                    // This is the original template that was requested so remember it to return.
+                                    templateSource = data.value;
+                                }
+                            });
+
+                            if (response.strings.length) {
+                                // If we have strings that the template needs then warm the string cache
+                                // with them now so that we don't need to re-fetch them.
+                                str.cache_strings(response.strings.map(function(data) {
+                                    return {
+                                        component: data.component,
+                                        key: data.name,
+                                        value: data.value
+                                    };
+                                }));
+                            }
+
+                            // Return the original template source that the user requested.
+                            return templateSource;
+                        });
+
+                        return templatePromises[searchKey];
+                    });
+            }
+
+            return promise
+                .then(function(source) {
+                    // When we've successfully loaded the template then resolve the deferred
+                    // in the buffer so that all of the calling code can proceed.
+                    return templateDeferred.resolve(source);
+                })
+                .catch(function(error) {
+                    // If there was an error loading the template then reject the deferred
+                    // in the buffer so that all of the calling code can proceed.
+                    templateDeferred.reject(error);
+                    // Rethrow for anyone else listening.
+                    throw error;
+                });
+        });
+
+        if (requests.length) {
+            // We have requests to send so resolve the deferred with the promises.
+            serverRequestsDeferred.resolve(ajax.call(requests, true, false, false, 0, M.cfg.templaterev));
+        } else {
+            // Nothing to load so we can resolve our deferred.
+            serverRequestsDeferred.resolve();
+        }
+
+        // Once we've finished loading all of the templates then recurse to process
+        // any templates that may have been added to the buffer in the time that we
+        // were fetching.
+        $.when.apply(null, templatePromises)
+            .then(function() {
+                // Remove the templates we've loaded from the buffer.
+                loadTemplateBuffer.splice(0, templatesToLoad.length);
+                isLoadingTemplates = false;
+                processLoadTemplateBuffer();
+                return;
+            })
+            .catch(function() {
+                // Remove the templates we've loaded from the buffer.
+                loadTemplateBuffer.splice(0, templatesToLoad.length);
+                isLoadingTemplates = false;
+                processLoadTemplateBuffer();
+            });
+    };
+
     /**
      * Constructor
      *
@@ -85,7 +283,7 @@ define([
     Renderer.prototype.currentThemeName = '';
 
     /**
-     * Load a template from the cache or local storage or ajax request.
+     * Load a template.
      *
      * @method getTemplate
      * @private
@@ -95,44 +293,44 @@ define([
      * @return {Promise} JQuery promise object resolved when the template has been fetched.
      */
     Renderer.prototype.getTemplate = function(templateName) {
+        var currentTheme = this.currentThemeName;
+        var searchKey = currentTheme + '/' + templateName;
+
+        // If we haven't already seen this template then buffer it.
+        var cachedPromise = getTemplatePromiseFromCache(searchKey);
+        if (cachedPromise) {
+            return cachedPromise;
+        }
+
+        // Check the buffer to seee if this template has already been added.
+        var existingBufferRecords = loadTemplateBuffer.filter(function(record) {
+            return record.searchKey == searchKey;
+        });
+        if (existingBufferRecords.length) {
+            // This template is already in the buffer so just return the existing
+            // promise. No need to add it to the buffer again.
+            return existingBufferRecords[0].deferred.promise();
+        }
+
+        // This is the first time this has been requested so let's add it to the buffer
+        // to be loaded.
         var parts = templateName.split('/');
         var component = parts.shift();
-        var name = parts.shift();
+        var name = parts.join('/');
+        var deferred = $.Deferred();
 
-        var searchKey = this.currentThemeName + '/' + templateName;
+        // Add this template to the buffer to be loaded.
+        loadTemplateBuffer.push({
+            component: component,
+            name: name,
+            theme: currentTheme,
+            searchKey: searchKey,
+            deferred: deferred
+        });
 
-        // First try request variables.
-        if (searchKey in templatePromises) {
-            return templatePromises[searchKey];
-        }
-
-        // Now try local storage.
-        var cached = storage.get('core_template/' + searchKey);
-
-        if (cached) {
-            templateCache[searchKey] = cached;
-            templatePromises[searchKey] = $.Deferred().resolve(cached).promise();
-            return templatePromises[searchKey];
-        }
-
-        // Oh well - load via ajax.
-        var promises = ajax.call([{
-            methodname: 'core_output_load_template',
-            args: {
-                component: component,
-                template: name,
-                themename: this.currentThemeName
-            }
-        }], true, false);
-
-        templatePromises[searchKey] = promises[0].then(
-            function(templateSource) {
-                templateCache[searchKey] = templateSource;
-                storage.set('core_template/' + searchKey, templateSource);
-                return templateSource;
-            }
-        );
-        return templatePromises[searchKey];
+        // We know there is at least one thing in the buffer so kick off a processing run.
+        processLoadTemplateBuffer();
+        return deferred.promise();
     };
 
     /**
@@ -266,6 +464,10 @@ define([
             param = parts.join(',').trim();
         }
 
+        if (!component || component === 'moodle') {
+            component = 'core';
+        }
+
         if (param !== '') {
             // Allow variable expansion in the param part only.
             param = helper(param, context);
@@ -298,8 +500,9 @@ define([
         // Escape the {{ and the ".
         // This involves wrapping {{, and }} in change delimeter tags.
         content = content
-            .replace('"', '\\"')
+            .replace(/"/g, '\\"')
             .replace(/([\{\}]{2,3})/g, '{{=<% %>=}}$1<%={{ }}=%>')
+            .replace(/(\r\n|\r|\n)/g, '&#x0a;')
             ;
         return '"' + content + '"';
     };
@@ -357,6 +560,60 @@ define([
         return '[[_t_' + index + ']]';
     };
 
+    /**
+     * Return a helper function to be added to the context for rendering the a
+     * template.
+     *
+     * This will parse the provided text before giving it to the helper function
+     * in order to remove any blacklisted nested helpers to prevent one helper
+     * from calling another.
+     *
+     * In particular to prevent the JS helper from being called from within another
+     * helper because it can lead to security issues when the JS portion is user
+     * provided.
+     *
+     * @param  {function} helperFunction The helper function to add
+     * @param  {object} context The template context for the helper function
+     * @return {Function} To be set in the context
+     */
+    Renderer.prototype.addHelperFunction = function(helperFunction, context) {
+        return function() {
+            return function(sectionText, helper) {
+                // Override the blacklisted helpers in the template context with
+                // a function that returns an empty string for use when executing
+                // other helpers. This is to prevent these helpers from being
+                // executed as part of the rendering of another helper in order to
+                // prevent any potential security issues.
+                var originalHelpers = blacklistedNestedHelpers.reduce(function(carry, name) {
+                    if (context.hasOwnProperty(name)) {
+                        carry[name] = context[name];
+                    }
+
+                    return carry;
+                }, {});
+
+                blacklistedNestedHelpers.forEach(function(helperName) {
+                    context[helperName] = function() {
+                        return '';
+                    };
+                });
+
+                // Execute the helper with the modified context that doesn't include
+                // the blacklisted nested helpers. This prevents the blacklisted
+                // helpers from being called from within other helpers.
+                var result = helperFunction.apply(this, [context, sectionText, helper]);
+
+                // Restore the original helper implementation in the context so that
+                // any further rendering has access to them again.
+                for (var name in originalHelpers) {
+                    context[name] = originalHelpers[name];
+                }
+
+                return result;
+            }.bind(this);
+        }.bind(this);
+    };
+
     /**
      * Add some common helper functions to all context objects passed to templates.
      * These helpers match exactly the helpers available in php.
@@ -371,24 +628,12 @@ define([
         this.requiredStrings = [];
         this.requiredJS = [];
         context.uniqid = (uniqInstances++);
-        context.str = function() {
-          return this.stringHelper.bind(this, context);
-        }.bind(this);
-        context.pix = function() {
-          return this.pixHelper.bind(this, context);
-        }.bind(this);
-        context.js = function() {
-          return this.jsHelper.bind(this, context);
-        }.bind(this);
-        context.quote = function() {
-          return this.quoteHelper.bind(this, context);
-        }.bind(this);
-        context.shortentext = function() {
-          return this.shortenTextHelper.bind(this, context);
-        }.bind(this);
-        context.userdate = function() {
-          return this.userDateHelper.bind(this, context);
-        }.bind(this);
+        context.str = this.addHelperFunction(this.stringHelper, context);
+        context.pix = this.addHelperFunction(this.pixHelper, context);
+        context.js = this.addHelperFunction(this.jsHelper, context);
+        context.quote = this.addHelperFunction(this.quoteHelper, context);
+        context.shortentext = this.addHelperFunction(this.shortenTextHelper, context);
+        context.userdate = this.addHelperFunction(this.userDateHelper, context);
         context.globals = {config: config};
         context.currentTheme = themeName;
     };
@@ -485,7 +730,7 @@ define([
      * content with the their respective translated dates.
      *
      * @param {String} content The content in which string placeholders are to be found.
-     * @param {Array} strings The strings to replace with.
+     * @param {Array} dates The dates to replace with.
      * @return {String} The treated content.
      */
     Renderer.prototype.treatDatesInContent = function(content, dates) {
@@ -808,6 +1053,33 @@ define([
             return renderer.render(templateName, context, themeName);
         },
 
+        /**
+         * Every call to render creates a new instance of the class and calls render on it. This
+         * means each render call has it's own class variables.
+         *
+         * This alernate to the standard .render() function returns the html and js in a single object suitable for a
+         * native Promise.
+         *
+         * @method renderForPromise
+         * @private
+         * @param {string} templateName - should consist of the component and the name of the template like this:
+         *                              core/menu (lib/templates/menu.mustache) or
+         *                              tool_bananas/yellow (admin/tool/bananas/templates/yellow.mustache)
+         * @param {Object} context - Could be array, string or simple value for the context of the template.
+         * @param {string} themeName - Name of the current theme.
+         * @return {Promise} JQuery promise object resolved when the template has been rendered.
+         */
+        renderForPromise: function(templateName, context, themeName) {
+            var renderer = new Renderer();
+            return renderer.render(templateName, context, themeName)
+            .then(function(html, js) {
+                return {
+                    html: html,
+                    js: js,
+                };
+            });
+        },
+
         /**
          * Every call to renderIcon creates a new instance of the class and calls renderIcon on it. This
          * means each render call has it's own class variables.
@@ -879,6 +1151,6 @@ define([
          */
         appendNodeContents: function(element, html, js) {
             domAppend(element, html, js);
-        }
+        },
     };
 });
